{-# OPTIONS -fglasgow-exts #-}

module Core (GCAT, GAT, GCU, 
             all_l, all_r, one_l, one_r,
             mother, xm,
             compose, compose', choice, choice', try, vsucc, vfail)
    where

import Data.Generics
import Maybe
import Monad

-- Type of a contextual accumulative transformer: 
type GCAT c a = forall t. Data t => c -> a -> t -> Maybe (a, t)
-- Type of a non-contextual accumulative transformer:
type GAT a  = forall t. Data t => a -> t -> Maybe (a, t)
-- Type of a context updater:
type GCU c = forall t. Typeable t => c -> t -> c

--------------------------------------------------------------------------------
-- One layer strategies:
--------------------------------------------------------------------------------
newtype Xall_l m a t = Xall_l { unXall_l :: m (a,t) }

all_l :: GAT a -> GAT a
all_l s a t = unXall_l (gfoldl k z t)
    where z d = Xall_l (return (a, d))
          k (Xall_l x) t = Xall_l (do (a,d) <- x
                                      (a',t') <- s a t
                                      return (a', d t'))


newtype Xall_r m a t = Xall_r { unXall_r :: a -> m (a,t) }

all_r :: GAT a -> GAT a
all_r s a t = unXall_r (gfoldl k z t) a
    where z d = Xall_r (\a -> return (a,d))
          k (Xall_r g) t = Xall_r (\a -> do (a',t') <- s a t
                                            (a'',d) <- g a'
                                            return (a'',d t'))

-- The second `t` is for when the first component is an mzero, it still has a
-- partially applied term. When the first component is not an mzero, the second
-- should not be used.
newtype Xone_l m a t = Xone_l { unXone_l :: (m (a,t),t)}

one_l :: GAT a -> GAT a
one_l s a t = fst (unXone_l (gfoldl k z t))
    where z d = Xone_l (mzero,d)
          k (Xone_l (x,d)) t = Xone_l (
            (do {(a,d) <- x; return (a,d t)} 
             `mplus`
             do {(a',t') <- s a t; return (a',d t')}),
            d t)
                          
newtype Xone_r m a t = Xone_r { unXone_r :: Bool -> (m (a,t),t)}

one_r :: GAT a -> GAT a
one_r s a t = fst (unXone_r (gfoldl k z t) False)
    where z d = Xone_r (\_ -> (mzero,d))
          k (Xone_r g) t = Xone_r 
            (\b -> (if b then mzero
                         else do (a',t') <- s a t
                                 return (a', snd (g True) t')
                              `mplus`
                              do (x,d) <- fst (g False)
                                 return (x,d t)
                   , snd (g True) t))

--------------------------------------------------------------------------------
-- Strategy combinators
--------------------------------------------------------------------------------

-- We don't want to operate on GCAT, but rather GAT, because generally multiple
-- visits might have different contexts.
-- note that 'compose f g' is equivalent to 'f ; g', not 'f . g'

-- These implementations look a little obscure, mainly due to that we want to
-- use the curried version of visits and traversals. Otherwise, there is really
-- a neat way of define all of them:

-- ; = (>>=)
-- + = mplus
-- try s = s + succ
-- repeat s = try (s ; (repeat s))
-- fail = mzero
-- succ = return

compose :: GAT a -> GAT a -> GAT a
compose s1 s2 a t = do (a',t') <- s1 a t
                       s2 a' t'

compose' :: GAT a -> GAT a -> GAT a
compose' s1 s2 = compose s2 s1

choice :: GAT a -> GAT a -> GAT a
choice s1 s2 a t = s1 a t `mplus` s2 a t

choice' :: GAT a -> GAT a -> GAT a
choice' s1 s2 = choice s2 s1

try :: GAT a -> GAT a
try s = choice s vsucc

rep :: GAT a -> GAT a
rep s = try (s `compose` rep s)

vfail :: GAT a
vfail _ _ = mzero

vsucc :: GAT a
vsucc = curry return

--------------------------------------------------------------------------------
-- Core recursive strategies
--------------------------------------------------------------------------------
-- full_td (s)  = s ; all (tull_td (s))
-- spine_td (s) = s ; one (spine_td (s))
-- stop_td (s)  = s + all (stop_td (s))
-- once_td (s)  = s + one (once_td (s))

-- bottom-up versions are just the flip side:

-- full_bu (s)  = all (tull_bu (s)) ; s
-- spine_bu (s) = one (spine_bu (s)) ; s
-- stop_bu (s)  = all (top_bu (s)) + s
-- once_bu (s)  = one (once_bu (s)) + s

-- Take this literally! Replace ; and + with flip (;) and flip (+)!

mother :: (GAT a -> GAT a -> GAT a) ->  -- {;,+}*{id,flip}
          (GAT a -> GAT a) ->           -- one-layer strategy {one,all}*{l,r} 
          GCU c -> 
          GCAT c a -> 
          GCAT c a
mother g f u s c a t = (s c `g` f (mother g f u s (u c t))) a t

-- fixpoint traversal innermost/outermost

-- innermost (s) = repeat (oncebu (s))
-- outermost (s) = repeat (oncetd (s))

xm :: (GAT a -> GAT a -> GAT a) -> -- {+, flip +}
      (GAT a -> GAT a) ->          -- {one_l, one_r}
      GCU c -> 
      GCAT c a -> 
      GCAT c a
xm g f u s c = rep (mother g f u s c)
